/*
	TuxPuck - boardentity.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include <math.h>
#include "video.h"
#include "audio.h"
#include "board.h"
#include "boardentity.h"
#include "global.h"

/* defines */
#define PAD_W			((float)9.284)
#define PAD_H			((float)0.02)
#define PUCK_W			((float)2.2)
#define ABS(x)			((x)<0 ? -(x) : (x))
#define MAX_PUCK_SPEED		((float)0.20)
#define THRESHOLD		((float)0.02)
#define MOVE_TO_SPEED		((float)0.05)
#define HIT_FRICTION		((float)0.60)

/* externals */
extern unsigned char pad_tga[];
extern unsigned char puck_tga[];
extern unsigned char nock_wav[];

/* BoardEntity implemention */
BoardEntity::BoardEntity(Board * board)
{
  _theBoard = board;
  _moveArea.x1 = -20;
  _moveArea.x2 = 20;
  _moveArea.z1 = -42;
  _moveArea.z2 = 42;
  _x = _z = _dx = _dz = _w = _h = 0;
  _y_scale = 1.0;
  _sdlImage = NULL;
  _transparent = 0;
  _oldRect.x = _oldRect.y = _oldRect.w = _oldRect.h = 0;
}

BoardEntity::~BoardEntity()
{
}

void BoardEntity::move(Uint32 time)
{
  _x += _dx * time;						     /* Add movement */
  _z += _dz * time;
  if (_x < _moveArea.x1 + _w / 2)
    {								     /* Check for boundaries */
      _x = _moveArea.x1 + _w / 2;
      _dx = -_dx;
    }
  else if (_x > _moveArea.x2 - _w / 2)
    {
      _x = _moveArea.x2 - _w / 2;
      _dx = -_dx;
    }
  if (_z < _moveArea.z1 + _h / 2)
    {
      _z = _moveArea.z1 + _h / 2;
      _dz = -_dz;
    }
  else if (_z > _moveArea.z2 - _h / 2)
    {
      _z = _moveArea.z2 - _h / 2;
      _dz = -_dz;
    }
}

void BoardEntity::blit(void)
{
  SDL_Surface *scaledSurface = NULL;
  SDL_Rect rect;
  float scale;

  scale = _theBoard->calcScale(_z);				     /* Calculate scale factor */
  scaledSurface = zoomSurface(_sdlImage, scale);		     /* Scale surface */
  rect.x = (Sint16) (320 - scaledSurface->w / 2 + scale * _x * 12.7);
  rect.y = (Sint16) (_theBoard->calcY(_z) - scaledSurface->h * _y_scale);
  if (_transparent)
    SDL_SetAlpha(scaledSurface, SDL_SRCALPHA, 128);
  else
    SDL_SetAlpha(scaledSurface, 0, 0);
  Video::instance()->blit(scaledSurface, &rect);
  _oldRect = rect;
  SDL_FreeSurface(scaledSurface);				     /* Delete the scaled surface */
}

void BoardEntity::erase(void)
{
  Video::instance()->erase(&_oldRect);
}

void BoardEntity::getPosition(float *x, float *z)
{
  if (x)
    *x = _x;
  if (z)
    *z = _z;
}

void BoardEntity::setVelocity(float dx, float dz)
{
  _dx = dx;
  _dz = dz;
}

void BoardEntity::getVelocity(float *dx, float *dz)
{
  if (dx)
    *dx = _dx;
  if (dz)
    *dz = _dz;
}

Uint8 BoardEntity::moveTowards(float x, float z, Uint32 time)
{
  float dx = 0, dz = 0;

  if (x > _x)
    {
      dx = MOVE_TO_SPEED;
      if (x < _x + dx * time)
	{
	  _x = x;
	  dx = 0;
	}
    }
  else if (x < _x)
    {
      dx = -MOVE_TO_SPEED;
      if (x > _x + dx * time)
	{
	  _x = x;
	  dx = 0;
	}
    }
  if (z > _z)
    {
      dz = MOVE_TO_SPEED;
      if (z < _z + dz * time)
	{
	  _z = z;
	  dz = 0;
	}
    }
  else if (z < _z)
    {
      dz = -MOVE_TO_SPEED;
      if (z > _z + dz * time)
	{
	  _z = z;
	  dz = 0;
	}
    }
  this->setVelocity(dx, dz);
  this->move(time);
  if (_x == x && _z == z)
    return 0;
  return 1;
}

void BoardEntity::toggleTransparent(void)
{
  if (_transparent)
    _transparent = 0;
  else
    _transparent = 1;
}

/* Pad implemention */
Pad::Pad(Board * board, Uint8 place):BoardEntity(board)
{
  _w = PAD_W;
  _h = PAD_H;
  _sdlImage = loadTga(pad_tga, NULL);
  SDL_SetColorKey(_sdlImage, SDL_SRCCOLORKEY, 2);
  if (place == 1)						     /* Player 1? */
    {
      _moveArea.z1 = -40;
      _moveArea.z2 = -20.8;
      _z = -36;
    }
  else								     /* Player 2 then */
    {
      _moveArea.z1 = 20.8;
      _moveArea.z2 = 40;
      _z = 36;
    }
  this->toggleTransparent();
}

Pad::~Pad()
{
  SDL_FreeSurface(_sdlImage);
}

/* Puck implemention */
Puck::Puck(Board * board):BoardEntity(board)
{
  _w = PUCK_W;
  _h = PUCK_W;
  _y_scale = 0.8;
  _sdlImage = loadTga(puck_tga, NULL);
  _state = PUCK_STATE_DEAD;
  _wavNock = Audio::instance()->createSound(nock_wav, NULL);
}

Puck::~Puck()
{
  SDL_FreeSurface(_sdlImage);
  Audio::instance()->freeSound(_wavNock);
}

bool Puck::willGetHitBy(Pad * pad, Uint32 time)
{
  float x, z, dx, dz;

  pad->getPosition(&x, &z);
  pad->getVelocity(&dx, &dz);
  if (ABS(x + dx * time - (_x + _dx * time)) > PAD_W / 2 + PUCK_W / 2)
    return false;
  if (z < _z)
    {
      if (z + dz * time + PAD_H / 2 > _z + _dz * time - PUCK_W / 2)
	return true;
      else
	return false;
    }
  else
    {
      if (z + dz * time - PAD_H / 2 < _z + _dz * time + PUCK_W / 2)
	return true;
      else
	return false;
    }
}

void Puck::setState(Uint8 state)
{
  _state = state;
}

void Puck::move(Uint32 time)
{
  if (_x + _dx * time < _moveArea.x1 + _w / 2
      || _x + _dx * time > _moveArea.x2 - _w / 2)
    Audio::instance()->playSound(_wavNock);
  ((BoardEntity *) this)->move(time);
}

bool Puck::isDead(void)
{
  if (_state == PUCK_STATE_DEAD)
    return true;
  else
    return false;
}

void Puck::getHitBy(Pad * pad, Uint32 time)
{
  float x, z, dx, dz, speed;

  pad->getPosition(&x, &z);
  pad->getVelocity(&dx, &dz);
  _dx += dx;
  _dx *= HIT_FRICTION;
  if (z < _z)
    {
      if (_dz < 0)
	{
	  _dz = -_dz;
	}
      _dz += dz;
      _dz *= HIT_FRICTION;
      _z = z + dz * time + PUCK_W / 2 + PAD_H / 2 + THRESHOLD;
    }
  else
    {
      if (_dz > 0)
	{
	  _dz = -_dz;
	}
      _dz += dz;
      _dz *= HIT_FRICTION;
      _z = z + dz * time - PUCK_W / 2 - PAD_H / 2 - THRESHOLD;
    }
  if (_dx * _dx + _dz * _dz > MAX_PUCK_SPEED * MAX_PUCK_SPEED)
    {
      speed = sqrt(_dx * _dx + _dz * _dz);
      _dx = _dx / speed * MAX_PUCK_SPEED;
      _dz = _dz / speed * MAX_PUCK_SPEED;
    }
  Audio::instance()->playSound(_wavNock);
}
